from bpy.types import Operator, Context, Object
from ..constants import MODERN_PRIMITIVE_PREFIX
from typing import cast, Iterable
from bpy.props import BoolProperty, EnumProperty
import bpy
from ..aux_func import (
    is_modern_primitive,
    calc_aabb,
    get_evaluated_vertices,
    mul_vert_mat,
)
from mathutils import Matrix, Vector, Quaternion
import math
import numpy as np
from ..aux_math import is_uniform


class BBox:
    def __init__(self, vert: Iterable[Vector]):
        (self.min, self.max) = calc_aabb(vert)
        self.size = self.max - self.min
        self.center = (self.min + self.max) / 2

    def __str__(self) -> str:
        return f"BBox(min={self.min}, max={self.max},\
size={self.size}, center={self.center})"


def _auto_axis(data: Iterable[Iterable[float]]) -> tuple[Vector, Vector, Vector]:
    # Converts the vertex coordinates to NumPy array
    verts = np.array(data)
    # Data standardization
    mean_coords = verts.mean(axis=0)
    verts = verts - mean_coords
    # calc Convariance matrix
    cov = np.cov(verts, rowvar=False)
    # Eigen values and Eigen vectors
    eigval, eigvec = np.linalg.eig(cov)
    # Sorting the unique vec in descending order
    sort_idx = eigval.argsort()[::-1]
    eigvec = eigvec[:, sort_idx]

    a0 = eigvec[:, :1].reshape(1, 3)
    a1 = eigvec[:, 1:2].reshape(1, 3)
    a2 = eigvec[:, 2:3].reshape(1, 3)
    return Vector(a0[0]), Vector(a1[0]), Vector(a2[0])


def to_4d_0(vec: Vector) -> Vector:
    ret = vec.to_4d()
    ret[3] = 0
    return ret


class ConvertTo_BaseOperator(Operator):
    @classmethod
    @property
    def type_name(cls):
        return cls.type.name

    @classmethod
    @property
    def bl_idname(cls):
        return f"mesh.{MODERN_PRIMITIVE_PREFIX}_convert_to_{cls.type_name.lower()}"

    @classmethod
    @property
    def bl_label(cls):
        return f"Convert object to Modern{cls.type_name}"

    bl_options = {"REGISTER", "UNDO"}
    keep_original: BoolProperty(name="Keep Original", default=False)
    apply_scale: BoolProperty(name="Apply Scaling", default=True)

    # The main axis of Height (the Width axis is the other)
    main_axis: EnumProperty(
        name="Base Axis",
        default="Auto",
        items=(
            ("Auto", "Auto", ""),
            ("X", "X", ""),
            ("Y", "Y", ""),
            ("Z", "Z", ""),
        ),
    )

    @classmethod
    def poll(cls, context: Context | None) -> bool:
        if context is None:
            return False
        context = cast(Context, context)

        sel = context.selected_objects
        if len(sel) == 0:
            return False
        for obj in sel:
            if obj is None or obj.type != "MESH":
                return False
        return True

    def _handle_proc(
        self,
        context: Context,
        obj: Object,
        bbox: BBox,
        mat: Matrix,
    ) -> tuple[Object, Vector]:
        raise NotImplementedError("This method should be implemented by subclass")

    def execute(self, context: Context | None) -> set[str]:
        sel = context.selected_objects.copy()
        # Copy the list because the object may be deleted in the loop
        for obj in sel.copy():
            # bound_box update
            if is_modern_primitive(obj):
                bpy.ops.object.mode_set(mode="OBJECT")
            obj.data.update()

            # Quotanion for rotating the main axis to the Z axis
            pre_rot: Quaternion
            # _handle Proc method handles the Z axis as height,
            #   so convert it in a timely manner.
            match self.main_axis:
                case "Auto":
                    # If the axis mode is Auto,
                    #   an error will be made if the scale value is not uniform
                    #        at this time.
                    if not is_uniform(obj.scale):
                        # If there is only one target object, treat it as an error
                        typ = "WARNING" if len(sel) > 1 else "ERROR"
                        self.report(
                            {typ},
                            f"Couldn't convert \"{obj.name}\" because It didn't have a uniform scaling value",  # noqa: E501
                        )
                        continue
                    axis = _auto_axis(get_evaluated_vertices(context, obj))
                    m = Matrix(
                        (
                            to_4d_0(axis[0]),
                            to_4d_0(axis[1]),
                            to_4d_0(axis[2]),
                            (0, 0, 0, 1),
                        )
                    )
                    _, pre_rot, _ = m.decompose()

                case "X":
                    # -90 degrees rotation around the Y axis
                    pre_rot = Quaternion(((0, 1, 0)), math.radians(-90))
                case "Y":
                    # 90 degrees around the X-axis
                    pre_rot = Quaternion((1, 0, 0), math.radians(90))
                case "Z":
                    # Do nothing
                    pre_rot = Quaternion()
            mat = pre_rot.to_matrix()

            # get bound_box info (size, average)
            # Bounding box when the z-axis is the main axis
            verts = mul_vert_mat(get_evaluated_vertices(context, obj), mat)
            bbox = BBox(verts)
            new_obj, offset = self._handle_proc(context, obj, bbox, mat)
            new_obj.name = obj.name + "_converted"

            new_obj.matrix_world = (
                obj.matrix_world
                @ pre_rot.inverted().to_matrix().to_4x4()
                @ Matrix.Translation(bbox.center)
                @ Matrix.Translation(offset)
            )

            if self.apply_scale:
                bpy.ops.object.mpr_apply_scale(strict=False)

            if not self.keep_original:
                bpy.data.objects.remove(obj)

        return {"FINISHED"}
